import socket
import select
import errno

"""Based on code from timeout_socket.py, with some tweaks for compatibility.
   These tweaks should really be rolled back into timeout_socket, but it's
   not totally clear who is maintaining it at this point. In the meantime,
   we'll use a different module name for our tweaked version to avoid any
   confusion.

   The original timeout_socket is by:

        Scott Cotton <scott@chronis.pobox.com>
        Lloyd Zusman <ljz@asfast.com>
        Phil Mayes <pmayes@olivebr.com>
        Piers Lauder <piers@cs.su.oz.au>
        Radovan Garabik <garabik@melkor.dnp.fmph.uniba.sk>
"""

ident = "$Id$"


WSAEINVAL = getattr(errno, 'WSAEINVAL', 10022)


class TimeoutSocket:

    """
    A socket imposter that supports timeout limits.
    """

    def __init__(self, timeout=20, sock=None):
        self.timeout = float(timeout)
        self.inbuf = ''
        if sock is None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock = sock
        self.sock.setblocking(0)
        self._rbuf = ''
        self._wbuf = ''

    def __getattr__(self, name):
        # Delegate to real socket attributes.
        return getattr(self.sock, name)

    def connect(self, *addr):
        timeout = self.timeout
        sock = self.sock
        try:
            # Non-blocking mode
            sock.setblocking(0)
            sock.connect(*addr)
            sock.setblocking(timeout != 0)
            return 1
        except socket.error as why:
            if not timeout:
                raise
            sock.setblocking(1)
            if len(why.args) == 1:
                code = 0
            else:
                code, why = why
            if code not in \
                    (errno.EINPROGRESS, errno.EALREADY, errno.EWOULDBLOCK):
                raise
            r, w, e = select.select([], [sock], [], timeout)
            if w:
                try:
                    sock.connect(*addr)
                    return 1
                except socket.error as why:
                    if len(why.args) == 1:
                        code = 0
                    else:
                        code, why = why
                    if code in (errno.EISCONN, WSAEINVAL):
                        return 1
                    raise
        raise TimeoutError('socket connect() timeout.')

    def send(self, data, flags=0):
        total = len(data)
        next = 0
        while 1:
            r, w, e = select.select([], [self.sock], [], self.timeout)
            if w:
                buff = data[next:next + 8192]
                sent = self.sock.send(buff, flags)
                next = next + sent
                if next == total:
                    return total
                continue
            raise TimeoutError('socket send() timeout.')

    def recv(self, amt, flags=0):
        if select.select([self.sock], [], [], self.timeout)[0]:
            return self.sock.recv(amt, flags)
        raise TimeoutError('socket recv() timeout.')

    buffsize = 4096
    handles = 1

    def makefile(self, mode="r", buffsize=-1):
        self.handles = self.handles + 1
        self.mode = mode
        return self

    def close(self):
        self.handles = self.handles - 1
        if self.handles == 0 and self.sock.fileno() >= 0:
            self.sock.close()

    def read(self, n=-1):
        if not isinstance(n, type(1)):
            n = -1
        if n >= 0:
            k = len(self._rbuf)
            if n <= k:
                data = self._rbuf[:n]
                self._rbuf = self._rbuf[n:]
                return data
            n = n - k
            L = [self._rbuf]
            self._rbuf = ""
            while n > 0:
                new = self.recv(max(n, self.buffsize))
                if not new:
                    break
                k = len(new)
                if k > n:
                    L.append(new[:n])
                    self._rbuf = new[n:]
                    break
                L.append(new)
                n = n - k
            return "".join(L)
        k = max(4096, self.buffsize)
        L = [self._rbuf]
        self._rbuf = ""
        while 1:
            new = self.recv(k)
            if not new:
                break
            L.append(new)
            k = min(k * 2, 1024 ** 2)
        return "".join(L)

    def readline(self, limit=-1):
        data = ""
        i = self._rbuf.find('\n')
        while i < 0 and not (0 < limit <= len(self._rbuf)):
            new = self.recv(self.buffsize)
            if not new:
                break
            i = new.find('\n')
            if i >= 0:
                i = i + len(self._rbuf)
            self._rbuf = self._rbuf + new
        if i < 0:
            i = len(self._rbuf)
        else:
            i = i + 1
        if 0 <= limit < len(self._rbuf):
            i = limit
        data, self._rbuf = self._rbuf[:i], self._rbuf[i:]
        return data

    def readlines(self, sizehint=0):
        total = 0
        list = []
        while 1:
            line = self.readline()
            if not line:
                break
            list.append(line)
            total += len(line)
            if sizehint and total >= sizehint:
                break
        return list

    def writelines(self, list):
        self.send(''.join(list))

    def write(self, data):
        self.send(data)

    def flush(self):
        pass


class TimeoutError(Exception):
    pass



# TODO add the license
# I had to rewrite this class because the python MIME email.mime (version 2.5)
# are buggy, they use \n instead \r\n for new line which is not compliant
# to standard!
# http://bugs.python.org/issue5525

# TODO do not load all the message in memory stream it from the disk

import re
import random
import sys


# new line
NL = '\r\n'

_width = len(repr(sys.maxint - 1))
_fmt = '%%0%dd' % _width


class MIMEMessage:

    def __init__(self):
        self._files = []
        self._xmlMessage = ""
        self._startCID = ""
        self._boundary = ""

    def makeBoundary(self):
        # create the boundary
        msgparts = []
        msgparts.append(self._xmlMessage)
        for i in self._files:
            msgparts.append(i.read())
        # this sucks, all in memory
        alltext = NL.join(msgparts)
        self._boundary = _make_boundary(alltext)
        # maybe I can save some memory
        del alltext
        del msgparts
        self._startCID = "<" + (_fmt % random.randrange(sys.maxint)) \
            + (_fmt % random.randrange(sys.maxint)) + ">"

    def toString(self):
        '''it return a string with the MIME message'''
        if len(self._boundary) == 0:
            # the makeBoundary hasn't been called yet
            self.makeBoundary()
        # ok we have everything let's start to spit the message out
        # first the XML
        returnstr = NL + "--" + self._boundary + NL
        returnstr += "Content-Type: text/xml; charset=\"us-ascii\"" + NL
        returnstr += "Content-Transfer-Encoding: 7bit" + NL
        returnstr += "Content-Id: " + self._startCID + NL + NL
        returnstr += self._xmlMessage + NL
        # then the files
        for file in self._files:
            returnstr += "--" + self._boundary + NL
            returnstr += "Content-Type: application/octet-stream" + NL
            returnstr += "Content-Transfer-Encoding: binary" + NL
            returnstr += "Content-Id: <" + str(id(file)) + ">" + NL + NL
            file.seek(0)
            returnstr += file.read() + NL
        # closing boundary
        returnstr += "--" + self._boundary + "--" + NL
        return returnstr

    def attachFile(self, file):
        '''
        it adds a file to this attachment
        '''
        self._files.append(file)

    def addXMLMessage(self, xmlMessage):
        '''
        it adds the XML message. we can have only one XML SOAP message
        '''
        self._xmlMessage = xmlMessage

    def getBoundary(self):
        '''
        this function returns the string used in the mime message as a
        boundary. First the write method as to be called
        '''
        return self._boundary

    def getStartCID(self):
        '''
        This function returns the CID of the XML message
        '''
        return self._startCID


def _make_boundary(text=None):
    # some code taken from python stdlib
    # Craft a random boundary.  If text is given, ensure that the chosen
    # boundary doesn't appear in the text.
    token = random.randrange(sys.maxint)
    boundary = ('=' * 10) + (_fmt % token) + '=='
    if text is None:
        return boundary
    b = boundary
    counter = 0
    while True:
        cre = re.compile('^--' + re.escape(b) + '(--)?$', re.MULTILINE)
        if not cre.search(text):
            break
        b = boundary + '.' + str(counter)
        counter += 1
    return b
